# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from ..core import Array


def _read_bound_key(infofile, ncpu):
    with open(infofile) as f:
        content = f.readlines()
    starting_line = None
    for num, line in enumerate(content):
        if len(set(["DOMAIN", "ind_min", "ind_max"]) - set(line.split())) == 0:
            starting_line = num + 1
            break
    bound_key = []
    if starting_line is not None:
        for n in range(ncpu):
            line = content[starting_line + n].split()
            bound_key.append(int(float(line[1])))
        bound_key.append(int(float(content[starting_line + ncpu - 1].split()[2])))
    return bound_key


def _btest(i, pos):
    return bool(i & (1 << pos))


def _hilbert3d(x, y, z, bit_length):

    i_bit_mask = np.zeros(3 * bit_length, dtype=bool)
    x_bit_mask = np.zeros(bit_length, dtype=bool)
    y_bit_mask = np.zeros_like(x_bit_mask)
    z_bit_mask = np.zeros_like(x_bit_mask)

    state_diagram = np.array([
        1, 2, 3, 2, 4, 5, 3, 5, 0, 1, 3, 2, 7, 6, 4, 5, 2, 6, 0, 7, 8, 8, 0, 7, 0, 7, 1,
        6, 3, 4, 2, 5, 0, 9, 10, 9, 1, 1, 11, 11, 0, 3, 7, 4, 1, 2, 6, 5, 6, 0, 6, 11,
        9, 0, 9, 8, 2, 3, 1, 0, 5, 4, 6, 7, 11, 11, 0, 7, 5, 9, 0, 7, 4, 3, 5, 2, 7, 0,
        6, 1, 4, 4, 8, 8, 0, 6, 10, 6, 6, 5, 1, 2, 7, 4, 0, 3, 5, 7, 5, 3, 1, 1, 11, 11,
        4, 7, 3, 0, 5, 6, 2, 1, 6, 1, 6, 10, 9, 4, 9, 10, 6, 7, 5, 4, 1, 0, 2, 3, 10, 3,
        1, 1, 10, 3, 5, 9, 2, 5, 3, 4, 1, 6, 0, 7, 4, 4, 8, 8, 2, 7, 2, 3, 2, 1, 5, 6,
        3, 0, 4, 7, 7, 2, 11, 2, 7, 5, 8, 5, 4, 5, 7, 6, 3, 2, 0, 1, 10, 3, 2, 6, 10, 3,
        4, 4, 6, 1, 7, 0, 5, 2, 4, 3
    ]).reshape((8, 2, 12), order='F')

    # Convert to binary
    for i in range(bit_length):
        x_bit_mask[i] = _btest(x, i)
        y_bit_mask[i] = _btest(y, i)
        z_bit_mask[i] = _btest(z, i)

    # Interleave bits
    for i in range(bit_length):
        i_bit_mask[3 * i + 2] = x_bit_mask[i]
        i_bit_mask[3 * i + 1] = y_bit_mask[i]
        i_bit_mask[3 * i] = z_bit_mask[i]

    # Build Hilbert ordering using state diagram
    cstate = 0
    for i in range(bit_length - 1, -1, -1):
        b2 = 0
        if i_bit_mask[3 * i + 2]:
            b2 = 1
        b1 = 0
        if i_bit_mask[3 * i + 1]:
            b1 = 1
        b0 = 0
        if i_bit_mask[3 * i]:
            b0 = 1
        sdigit = b2 * 4 + b1 * 2 + b0
        nstate = state_diagram[sdigit, 0, cstate]
        hdigit = state_diagram[sdigit, 1, cstate]
        i_bit_mask[3 * i + 2] = _btest(hdigit, 2)
        i_bit_mask[3 * i + 1] = _btest(hdigit, 1)
        i_bit_mask[3 * i] = _btest(hdigit, 0)
        cstate = nstate

    order = 0
    for i in range(3 * bit_length):
        b0 = 0
        if i_bit_mask[i]:
            b0 = 1
        order = order + b0 * (2**i)
    return order


def _get_cpu_list(bounding_box, lmax, levelmax, infofile, ncpu, ndim):

    bound_key = _read_bound_key(infofile=infofile, ncpu=ncpu)

    xmin = bounding_box["xmin"]
    xmax = bounding_box["xmax"]
    ymin = bounding_box["ymin"]
    ymax = bounding_box["ymax"]
    zmin = bounding_box["zmin"]
    zmax = bounding_box["zmax"]
    dmax = max(xmax - xmin, ymax - ymin, zmax - zmin)
    for ilevel in range(1, lmax + 1):
        dx = 0.5**ilevel
        if dx < dmax:
            break

    lmin = ilevel
    bit_length = lmin - 1
    maxdom = 2**bit_length
    imin = 0
    imax = 0
    jmin = 0
    jmax = 0
    kmin = 0
    kmax = 0
    if bit_length > 0:
        imin = int(xmin * maxdom)
        imax = imin + 1
        jmin = int(ymin * maxdom)
        jmax = jmin + 1
        kmin = int(zmin * maxdom)
        kmax = kmin + 1

    dkey = (2**(levelmax + 1) // maxdom)**ndim
    ndom = 1
    if bit_length > 0:
        ndom = 8
    idom = [imin, imax] * 4
    jdom = [jmin, jmin, jmax, jmax] * 2
    kdom = [kmin] * 4 + [kmax] * 4

    bounding_min = [0, 0, 0, 0, 0, 0, 0, 0]
    bounding_max = [0, 0, 0, 0, 0, 0, 0, 0]
    bounding = None
    for i in range(ndom):
        if bit_length > 0:
            bounding = _hilbert3d(idom[i], jdom[i], kdom[i], bit_length)
            order_min = bounding
        else:
            order_min = 0
        bounding_min[i] = order_min * dkey
        bounding_max[i] = (order_min + 1) * dkey

    cpu_min = [0, 0, 0, 0, 0, 0, 0, 0]
    cpu_max = [0, 0, 0, 0, 0, 0, 0, 0]
    for impi in range(ncpu):
        for i in range(ndom):
            if ((bound_key[impi] <= bounding_min[i])
                    and (bound_key[impi + 1] > bounding_min[i])):
                cpu_min[i] = impi

            if ((bound_key[impi] < bounding_max[i])
                    and (bound_key[impi + 1] >= bounding_max[i])):
                cpu_max[i] = impi

    cpu_list = []
    for i in range(ndom):
        for j in range(cpu_min[i], cpu_max[i] + 1):
            if j + 1 not in cpu_list:
                cpu_list.append(j + 1)

    return cpu_list


def hilbert_cpu_list(meta, scaling, select, infofile):
    if meta["ordering type"] != "hilbert":
        return
    bounding_box = {"xmin": 0, "xmax": 1, "ymin": 0, "ymax": 1, "zmin": 0, "zmax": 1}
    # Make an array of cell centers according to lmax
    box_size = (meta["boxlen"] * scaling).magnitude
    ncells = 2**min(meta["levelmax"], 18)  # limit to 262000 cells
    half_dxmin = 0.5 * box_size / ncells
    xyz_centers = Array(values=np.linspace(half_dxmin, box_size - half_dxmin, ncells),
                        unit=1.0 * scaling.units)
    new_bbox = False
    for c in "xyz":
        if c in select:
            new_bbox = True
            func_test = select[c](xyz_centers)
            inds = np.argwhere(func_test).ravel()
            start = xyz_centers[inds.min()] - (half_dxmin * scaling.units)
            end = xyz_centers[inds.max()] + (half_dxmin * scaling.units)
            bounding_box["{}min".format(c)] = start._array / box_size
            bounding_box["{}max".format(c)] = end._array / box_size
            select["xyz_{}".format(c)] = select.pop(c)

    if new_bbox:
        return _get_cpu_list(bounding_box=bounding_box,
                             lmax=meta["lmax"],
                             levelmax=meta["levelmax"],
                             infofile=infofile,
                             ncpu=meta["ncpu"],
                             ndim=meta["ndim"])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
import numpy as np
from .. import config
from ..core import Array
from pint.quantity import Quantity


def quiver(ax, x, y, z, density=1, color="w", **kwargs):
    """
    Wrapper around Matplotlib's quiver plot.

    We have a special argument `density` which aims to control the density of
    arrows in a similar manner to the density of streamlines in the streamplot.
    Matplotlib's own way of controlling the density of arrows is not so
    obvious.
    """
    default_args = {
        "angles": "xy",
        "pivot": "mid",
    }
    default_args.update(kwargs)

    skips = np.around(np.array(z.shape) * 4.0 / 128.0 / density).astype(np.int)
    skip = (slice(None, None, skips[0]), slice(None, None, skips[1]))

    args = [x[skip[0]], y[skip[1]], z[..., 0][skip], z[..., 1][skip]]
    if isinstance(color, str):
        default_args["color"] = color
    else:
        args.append(z[..., 2][skip])

    return ax.quiver(*args, **default_args)


def pcolormesh(ax, x, y, z, **kwargs):
    """
    Wrapper around Matplotlib's pcolormesh plot.
    """
    default_args = {
        "shading": "nearest",
    }
    default_args.update(kwargs)
    if "cmap" not in kwargs:
        kwargs["cmap"] = config.parameters["cmap"]
    return ax.pcolormesh(x, y, z, **default_args)


def contour(ax, x, y, z, labels=True, **kwargs):
    """
    Wrapper around Matplotlib's contour plot.

    We add a small convenience argument `labels` that allows to simply add
    clabels with a simple `labels=True`.
    """
    cs = ax.contour(x, y, z, **kwargs)
    if labels:
        ax.clabel(cs, inline=1, fontsize=10)
    return cs


def contourf(ax, x, y, z, **kwargs):
    """
    Wrapper around Matplotlib's contourf plot.
    """
    if "cmap" not in kwargs:
        kwargs["cmap"] = config.parameters["cmap"]
    return ax.contourf(x, y, z, **kwargs)


def streamplot(ax, x, y, z, **kwargs):
    """
    Wrapper around Matplotlib's streamplot plot.
    """
    default_args = {"color": "w"}
    default_args.update(kwargs)
    return ax.streamplot(x, y, z[..., 0], z[..., 1], **default_args)


def scatter(ax, x, y, data, **kwargs):
    """
    Wrapper around Matplotlib's scatter plot.
    If a point size has a unit, use PatchCollection instead of scatter.
    We use the scatter API for the arguments.
    If PatchCollection is used, we convert the scatter args to the
    PatchCollection syntax (e.g. "c" -> "color").
    """
    default_args = {"c": "b", "edgecolors": "k"}
    default_args.update(kwargs)
    use_patchcollection = False
    if "s" in default_args:
        if isinstance(default_args["s"], Array):
            default_args["s"] = default_args["s"].norm.values
            use_patchcollection = True
        if isinstance(default_args["s"], Quantity):
            default_args["s"] = np.full_like(x, default_args["s"].magnitude)
            use_patchcollection = True
    if use_patchcollection:
        array = None
        norm = None
        if "c" in default_args:
            if "facecolors" not in default_args:
                if isinstance(default_args["c"], str):
                    default_args["facecolors"] = default_args["c"]
                else:
                    array = default_args["c"]
            del default_args["c"]
        if "norm" in default_args:
            norm = default_args["norm"]
            del default_args["norm"]
        try:
            iter(default_args["s"])
        except TypeError:
            default_args["s"] = [default_args["s"]] * len(x)
        patches = [
            plt.Circle([x_, y_], s) for x_, y_, s in zip(x, y, default_args["s"])
        ]
        del default_args["s"]
        coll = ax.add_collection(PatchCollection(patches, **default_args))
        if array is not None:
            coll.set_array(array)
        if norm is not None:
            coll.set_norm(norm)
        return coll
    else:
        return ax.scatter(x, y, **default_args)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from ..core.tools import perpendicular_vector
from ..core import Array


def get_slice_direction(direction=None, dataset=None, dx=None, dy=None, origin=None):
    """
    Find direction vectors for slice.

    Direction can be:


    The origin can be either a vector of 3 numbers (xyz), or it can be "sink17"
    for sink particles.
    """

    # # Transform origin to coordinates if sink is requested
    # try:
    #     if origin.startswith("sink"):
    #         isink = np.where(holder.sinks["id"] ==
    #          int(origin.split(":")[1]))[0][0]
    #         origin = [holder.sinks["x"][isink], holder.sinks["y"]
    #                   [isink], holder.sinks["z"][isink]]
    # except AttributeError:
    #     pass

    ndim = dataset.meta["ndim"]

    dir_list = {"x": [1, 0, 0], "y": [0, 1, 0], "z": [0, 0, 1]}
    # dir_type = len(np.shape(direction))

    if ndim < 3:
        dir_vecs = np.array([[0., 0.], [1., 0.], [0., 1.]])

    elif direction in ["auto", "top", "side"]:
        if dx is None or dy is None:
            raise RuntimeError("When using automatic slice orientation, "
                               "dx cannot be None.")
        sphere_rad = 0.25 * (dx + dy)
        xyz = dataset["amr"]["xyz"] - origin
        # Compute angular momentum vector
        sphere = np.where(xyz.norm < sphere_rad.magnitude)
        pos = xyz * dataset["hydro"]["mass"]
        vel = dataset["hydro"]["velocity"]

        AngMom = np.sum(np.cross(pos.array[sphere], vel.array[sphere]), axis=0)
        if direction == "side":
            # Choose a vector perpendicular to the angular momentum vector
            dir3 = AngMom
            dir1 = perpendicular_vector(dir3)
            dir2 = np.cross(dir1, dir3)
        else:
            dir1 = AngMom
            dir2 = perpendicular_vector(dir1)
            dir3 = np.cross(dir1, dir2)
        # elif view ==
        # else:
        #     raise ValueError("Unknown view direction.")
        norm1 = np.linalg.norm(dir1)
        print("Normal slice vector: [%.5e,%.5e,%.5e]" %
              (dir1[0] / norm1, dir1[1] / norm1, dir1[2] / norm1))
        dir_vecs = np.array([dir1, dir2, dir3])

    elif isinstance(direction, str):
        if len(direction) == 3:  # This is the case where direction = "xyz"
            dir_vecs = np.array([
                dir_list[direction[0]], dir_list[direction[1]], dir_list[direction[2]]
            ])
        elif direction == "x":
            dir_vecs = np.array([dir_list["x"], dir_list["y"], dir_list["z"]])
        elif direction == "y":
            dir_vecs = np.array([dir_list["y"], dir_list["z"], dir_list["x"]])
        elif direction == "z":
            dir_vecs = np.array([dir_list["z"], dir_list["x"], dir_list["y"]])
    # This is the case where direction = [1,1,2]
    # (i.e. is a vector with 3 numbers)
    elif len(direction) == 3:
        dir1 = direction
        dir2 = perpendicular_vector(dir1)
        dir3 = np.cross(dir1, dir2).tolist()
        dir_vecs = np.array([dir1, dir2, dir3])
    # This is the case where two vectors are specified:
    # direction = [[1,0,1],[0,1,0]]
    elif len(direction) == 2:
        dir_vecs = np.array(
            [direction[0], direction[1],
             np.cross(direction[0], direction[1])])
    else:
        print("Bad direction for slice: ", direction)
        return

    # Avoid division by zero in norm
    norm = np.linalg.norm(dir_vecs, axis=1).reshape(3, 1)
    norm[norm == 0.] = 1.0
    dir_vecs = dir_vecs / norm

    if origin is None:
        origin = Array(values=np.zeros([1, ndim]), unit=dataset["amr"]["xyz"].unit)

    return dir_vecs, origin



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

# flake8: noqa

from .config import config, units
from .plot import histogram, plane, scatter
from .core import Array, Datagroup, Dataset, Plot



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

from matplotlib.colors import LogNorm, Normalize


def get_norm(norm=None, vmin=None, vmax=None):
    if norm == "log":
        return LogNorm(vmin=vmin, vmax=vmax)
    else:
        return Normalize(vmin=vmin, vmax=vmax)


def parse_layer(layer,
                mode=None,
                norm=None,
                vmin=None,
                vmax=None,
                operation=None,
                **kwargs):

    if isinstance(layer, dict):
        params = {
            key: layer[key]
            for key in set(layer.keys()) - set(["data", "mode", "operation"])
        }
        if "norm" not in params:
            params["norm"] = norm
        if "vmin" in params:
            vmin = params["vmin"]
            del params["vmin"]
        if "vmax" in params:
            vmax = params["vmax"]
            del params["vmax"]

        params["norm"] = get_norm(norm=params["norm"], vmin=vmin, vmax=vmax)

        for key, arg in kwargs.items():
            if key not in params:
                params[key] = arg

        settings = {}
        for key in ["mode", "operation"]:
            settings[key] = layer[key] if key in layer else eval(key)
        return layer["data"], settings, params
    else:
        params = {"norm": get_norm(norm=norm, vmin=vmin, vmax=vmax)}
        settings = {"mode": mode, "operation": operation}
        params.update(kwargs)
        return layer, settings, params



import numpy as np
from . import utils
from ..core import Array
from enum import Enum


class ReaderKind(Enum):
    AMR = 0
    SINK = 1
    PART = 2


class Reader():
    def __init__(self, kind=None):
        self.variables = {}
        self.offsets = {}
        self.meta = {}
        self.bytes = None
        self.initialized = False
        self.kind = kind

    def allocate_buffers(self, ngridmax, twotondim):
        for item in self.variables.values():
            if item["read"]:
                item["buffer"] = Array(values=np.zeros([ngridmax, twotondim],
                                                       dtype=np.dtype(item["type"])),
                                       unit=1.0 * item["unit"].units)

    def read_header(self, *args, **kwargs):
        return

    def read_level_header(self, *args, **kwargs):
        return

    def read_domain_header(self, *args, **kwargs):
        return

    def read_cacheline_header(self, *args, **kwargs):
        return

    def read_variables(self, ncache, ind, ilevel, cpuid, info):
        for item in self.variables.values():
            if item["read"]:
                item["buffer"]._array[:ncache, ind] = np.array(
                    utils.read_binary_data(
                        fmt="{}{}".format(ncache, item["type"]),
                        content=self.bytes,
                        offsets=self.offsets)) * item["unit"].magnitude
            else:
                self.offsets[item["type"]] += ncache
                self.offsets["n"] += 1

    def make_conditions(self, select, ncache):
        conditions = {}
        if not isinstance(select, bool):
            for key, func in select.items():
                if not isinstance(func, bool):
                    if key in self.variables:
                        conditions[key] = func(
                            self.variables[key]["buffer"][:ncache, :])
        return conditions

    def read_footer(self, *args, **kwargs):
        return

    def step_over(self, ncache, twotondim, ndim):
        self.offsets['d'] += ncache * twotondim * len(self.variables)
        self.offsets['n'] += twotondim * len(self.variables)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

from pint.quantity import Quantity
from ..core import Plot, Array
from .. import units
from .render import render
from .parser import parse_layer


def scatter(x,
            y,
            ax=None,
            logx=False,
            logy=False,
            loglog=False,
            norm=None,
            filename=None,
            title=None,
            xmin=None,
            xmax=None,
            ymin=None,
            ymax=None,
            vmin=None,
            vmax=None,
            **kwargs):
    """
    Plot a 2D scatter plot with two variables as input.
    """
    if loglog:
        logx = logy = True

    xvals = x.norm.values
    yvals = y.norm.values

    _, _, params = parse_layer(layer=None, norm=norm, vmin=vmin, vmax=vmax, **kwargs)
    to_render = [{
        "data": None,
        "mode": "scatter",
        "params": params,
    }]
    if "c" in params and not isinstance(params["c"], str):
        to_render[0].update({"unit": params["c"].unit.units, "name": params["c"].name})
        params["c"] = params["c"].norm.values
    if "s" in params:
        unit = None
        if isinstance(params["s"], Array):
            unit = params["s"].unit.units
        if isinstance(params["s"], Quantity):
            unit = params["s"].units
        if unit is not None:
            if unit != units.dimensionless:
                if x.unit.units != y.unit.units:
                    raise RuntimeError("Scatter: an Array with a unit was supplied "
                                       "as a size, but the units of the x and y "
                                       "Arrays do not agree. The size must either "
                                       "be a float or a dimensionless Array.")
                params["s"] = params["s"].to(x.unit.units)

    figure = render(x=xvals, y=yvals, data=to_render, logx=logx, logy=logy, ax=ax)

    figure["ax"].set_xlabel(x.label)
    figure["ax"].set_ylabel(y.label)
    figure["ax"].set_xlim(xmin, xmax)
    figure["ax"].set_ylim(ymin, ymax)

    return Plot(x=xvals, y=yvals, layers=to_render, fig=figure["fig"], ax=figure["ax"])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from .. import units


def get_finite_inds(x):
    """
    Find indices of finite numbers.
    """
    return np.where(np.isfinite(x))[0]


def finmin(x):
    """
    Finite minimum.
    """
    return np.amin(x.take(get_finite_inds(x)))


def finmax(x):
    """
    Finite maximum.
    """
    return np.amax(x.take(get_finite_inds(x)))


def to_bin_centers(x):
    """
    Convert array edges to centers
    """
    return 0.5 * (x[1:] + x[:-1])


def to_bin_edges(x):
    """
    Convert array centers to edges
    """
    centers = to_bin_centers(x)
    left = centers[0] - (x[1] - x[0])
    right = centers[-1] + (x[-1] - x[-2])
    return np.concatenate(np.concatenate(left, centers), right)


def perpendicular_vector(v):
    """
    Compute a vector perpendicular to the input vector
    """

    # x = y = z = 0 is not an acceptable solution
    if v[0] == v[1] == v[2] == 0:
        raise ValueError("zero-vector")

    if v[2] == 0:
        return [-v[1], v[0], 0]
    else:
        return [1.0, 1.0, -1.0 * (v[0] + v[1]) / v[2]]


def value_to_string(val, precision=3):
    """
    Convert a number to a human readable string.
    """
    if (not isinstance(val, float)) or (val == 0):
        text = str(val)
    elif (abs(val) >= 10.0**(precision)) or \
         (abs(val) <= 10.0**(-precision)):
        text = "{val:.{prec}e}".format(val=val, prec=precision)
    else:
        text = "{}".format(val)
        if len(text) > precision + 2 + (text[0] == '-'):
            text = "{val:.{prec}f}".format(val=val, prec=precision)
    return text


def make_label(name=None, unit=None):
    """
    Make an axis label from a name and a unit.
    """
    lab = ""
    if name:
        lab += name
    if unit and unit != units.dimensionless:
        if name:
            lab += " "
        lab += "[{:~}]".format(unit)
    return lab.strip()


def apply_mask(array, fill=np.nan):
    """
    Apply the mask on a numpy array if it exists, converting the masked values
    to the `fill` value (np.nan by default).
    """
    if hasattr(array, "mask"):
        return array.filled(fill)
    return array


def bytes_to_human_readable(size):
    multipliers = {"G": 1.0e9, "M": 1.0e6, "K": 1.0e3, "B": 1.0}
    for m, mult in multipliers.items():
        if size >= mult:
            return "{:.2f} {}B".format(size / mult, m)
    return "0B"



import osyris
import numpy as np


def test_1d_array():
    a = np.arange(100.)
    array = osyris.Array(values=a, unit="m")
    assert array.unit == osyris.units('m')
    assert len(array) == len(a)
    assert array.shape == a.shape
    assert np.allclose(array.values, a)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
import numpy.ma as ma
from pint.quantity import Quantity
from .slice import get_slice_direction
from .render import render
from .scatter import scatter
from .parser import parse_layer
from ..core import Plot, Array
from ..core.tools import to_bin_centers, apply_mask
from scipy.stats import binned_statistic_2d


def _add_scatter(datax, datay, to_scatter, origin, datadx, dir_vecs, dx, dy, ax):
    xyz = to_scatter[0]["data"] - origin
    viewport = max(dx.magnitude, dy.magnitude)
    radius = None
    if "s" in to_scatter[0]["params"]:
        size = to_scatter[0]["params"]["s"]
        if isinstance(size, Array) or isinstance(size, Quantity):
            radius = size.to(datax.unit.units)
            to_scatter[0]["params"]["s"] = radius
    if radius is None:
        # Fudge factor to select sinks close to the plane
        radius = Array(values=viewport * 0.05, unit=dx.units)
    dist1 = np.sum(xyz * dir_vecs[0], axis=1)
    global_selection = np.arange(len(to_scatter[0]["data"]))
    select = np.ravel(np.where(np.abs(dist1) <= radius))
    global_selection = global_selection[select]
    if len(select) > 0:
        # Project coordinates onto the plane by taking dot product with axes vectors
        coords = xyz[select]
        datax = np.inner(coords, dir_vecs[1])
        datay = np.inner(coords, dir_vecs[2])
        if dx is not None:
            # Limit selection further by using distance from center
            dist2 = coords
            select2 = np.ravel(
                np.where(np.abs(dist2.norm.values) <= viewport * 0.6 * np.sqrt(2.0)))
            datax = datax[select2]
            datay = datay[select2]
            global_selection = global_selection[select2]
        if "c" in to_scatter[0]["params"]:
            # TODO: also check that parents are the same to ensure size match?
            if isinstance(to_scatter[0]["params"]["c"], Array):
                to_scatter[0]["params"]["c"] = to_scatter[0]["params"]["c"][
                    global_selection]
        scatter(x=datax, y=datay, ax=ax, **to_scatter[0]["params"])


def plane(*layers,
          direction="z",
          dx=None,
          dy=None,
          fname=None,
          title=None,
          plot=True,
          mode=None,
          norm=None,
          vmin=None,
          vmax=None,
          operation="mean",
          origin=None,
          resolution=256,
          ax=None,
          **kwargs):
    """
    Plot a 2D slice through the data domain.
    """

    if isinstance(layers, Array):
        layers = [layers]

    to_process = []
    to_render = []
    operations = []
    to_scatter = []
    for layer in layers:
        data, settings, params = parse_layer(layer=layer,
                                             mode=mode,
                                             norm=norm,
                                             vmin=vmin,
                                             vmax=vmax,
                                             operation=operation,
                                             **kwargs)
        if settings["mode"] == "scatter":
            to_scatter.append({"data": data, "params": params})
        else:
            to_process.append(data)
            to_render.append({
                "mode": settings["mode"],
                "params": params,
                "unit": data.unit.units,
                "name": data.name
            })
            operations.append(settings["operation"])

    dataset = to_process[0].parent.parent

    # Set window size
    if dy is None:
        dy = dx
    if dx is not None and not isinstance(dx, Quantity):
        dx *= dataset["amr"]["xyz"].unit
    if dy is not None and not isinstance(dy, Quantity):
        dy *= dataset["amr"]["xyz"].unit

    dir_vecs, origin = get_slice_direction(direction=direction,
                                           dataset=dataset,
                                           dx=dx,
                                           dy=dy,
                                           origin=origin)

    # Distance to the plane
    xyz = dataset["amr"]["xyz"] - origin
    diagonal = dataset["amr"]["dx"] * np.sqrt(dataset.meta["ndim"]) * 0.5
    dist1 = np.sum(xyz * dir_vecs[0], axis=1)
    # Create an array of indices to allow further narrowing of the selection below
    global_selection = np.arange(len(dataset["amr"]["dx"]))

    # Select cells in contact with plane
    select = np.ravel(np.where(np.abs(dist1) <= diagonal))
    global_selection = global_selection[select]

    if len(select) == 0:
        raise RuntimeError("No cells were selected to construct the plane. "
                           "The resulting figure would be empty.")

    # Project coordinates onto the plane by taking dot product with axes vectors
    coords = xyz[select]
    datax = np.inner(coords, dir_vecs[1])
    datay = np.inner(coords, dir_vecs[2])
    datadx = 0.5 * dataset["amr"]["dx"][select]

    # Get limits
    limits = {
        'xmin': np.amin(datax - datadx).values,
        'xmax': np.amax(datax + datadx).values,
        'ymin': np.amin(datay - datadx).values,
        'ymax': np.amax(datay + datadx).values
    }

    # Define slice extent
    if dx is None:
        xmin = limits['xmin']
        xmax = limits['xmax']
        ymin = limits['ymin']
        ymax = limits['ymax']
    else:
        xmin = -0.5 * dx.magnitude
        xmax = xmin + dx.magnitude
        ymin = -0.5 * dy.magnitude
        ymax = ymin + dy.magnitude
        # Limit selection further by using distance from center
        dist2 = coords - datadx * np.sqrt(dataset.meta["ndim"])
        select2 = np.ravel(
            np.where(
                np.abs(dist2.norm.values) <= max(dx.magnitude, dy.magnitude) * 0.6 *
                np.sqrt(2.0)))
        coords = coords[select2]
        datax = datax[select2]
        datay = datay[select2]
        datadx = datadx[select2]
        global_selection = global_selection[select2]

    scalar_layer = []
    to_binning = []
    for ind in range(len(to_process)):
        if to_render[ind]["mode"] in ["vec", "stream"]:
            if to_process[ind].ndim < 3:
                uv = to_process[ind].array[global_selection]
            else:
                uv = np.inner(to_process[ind].array.take(global_selection, axis=0),
                              dir_vecs[1:])
            w = None
            if "color" in to_render[ind]["params"]:
                if isinstance(to_render[ind]["params"]["color"], Array):
                    w = to_render[ind]["params"]["color"].norm.values
                elif isinstance(to_render[ind]["params"]["color"], np.ndarray):
                    w = to_render[ind]["params"]["color"]
            if w is None:
                w = np.linalg.norm(uv, axis=1)
            else:
                w = w.take(global_selection, axis=0)
            to_binning.append(apply_mask(uv[:, 0]))
            to_binning.append(apply_mask(uv[:, 1]))
            to_binning.append(apply_mask(w))
            scalar_layer.append(False)
        else:
            to_binning.append(apply_mask(to_process[ind].norm.values[global_selection]))
            scalar_layer.append(True)

    # Buffer for counts
    to_binning.append(np.ones_like(to_binning[0]))

    # Construct some bin edges
    if isinstance(resolution, int):
        resolution = {'x': resolution, 'y': resolution}
    xedges = np.linspace(xmin, xmax, resolution['x'] + 1)
    yedges = np.linspace(ymin, ymax, resolution['y'] + 1)

    # In the contour plots, x and y are the centers of the cells, instead of the edges
    xcenters = to_bin_centers(xedges)
    ycenters = to_bin_centers(yedges)

    # First histogram the cell centers into the grid bins
    binned, _, _, _ = binned_statistic_2d(x=apply_mask(datay.array),
                                          y=apply_mask(datax.array),
                                          values=to_binning,
                                          statistic="mean",
                                          bins=[yedges, xedges])

    # Next, find all the empty pixels, find the cell is lies in a apply the value
    condition = np.isnan(binned[-1])
    xgrid, ygrid = np.meshgrid(xcenters, ycenters, indexing='xy')
    # Make array of cell indices (in original array of cells) with image shape
    indices = np.zeros_like(binned[-1], dtype=int)
    # We also need a mask for pixels that find no cells (outside of the data range)
    mask = np.zeros_like(binned[-1], dtype=bool)

    pixel_positions = xgrid.reshape(xgrid.shape + (1, )) * dir_vecs[1] + ygrid.reshape(
        ygrid.shape + (1, )) * dir_vecs[2]
    # We only need to search in the cells above a certain size
    large_cells = np.ravel(
        np.where(datadx >= 0.25 * (min(xedges[1] - xedges[0], yedges[1] - yedges[0]))))
    coords = coords[large_cells]
    large_cells_dx = datadx.array[large_cells]
    global_indices = np.arange(len(datadx))[large_cells]

    # To keep memory usage down to a minimum, we process the image one column at a time
    for i in range(indices.shape[-1]):
        # We know we are looking only at a column of cells, so we make a line from the
        # two end points (x1, x2), compute distance to the line to filter out the cells
        x1 = pixel_positions[0, i, :]
        x2 = pixel_positions[-1, i, :]
        x0_minus_x1 = coords.array - x1
        x0_minus_x2 = coords.array - x2
        x2_minus_x1 = x2 - x1
        distance_to_line = np.cross(x0_minus_x1, x0_minus_x2)
        if distance_to_line.ndim > 1:
            distance_to_line = np.linalg.norm(distance_to_line, axis=-1)
        else:
            distance_to_line = np.abs(distance_to_line)
        distance_to_line /= np.linalg.norm(x2_minus_x1)
        column = np.ravel(np.where(distance_to_line <= np.sqrt(3.0) * large_cells_dx))

        if len(column) > 0:
            distance_to_cell = []
            for n, c in enumerate("xyz"[:xyz.ndim]):
                distance_to_cell.append(pixel_positions[..., i, n:n + 1] -
                                        getattr(coords, c).array[column])
            # Find the cell where the x, y, z distance is smaller than dx/2
            inds = np.logical_and.reduce(
                [np.abs(d) <= large_cells_dx[column] for d in distance_to_cell])
            index_found = inds.max(axis=-1)
            index_value = global_indices[column][inds.argmax(axis=-1)]
            indices[:, i][index_found] = index_value[index_found]
            mask[:, i][np.logical_and(~index_found, condition[:, i])] = True
        else:
            mask[:, i][condition[:, i]] = True

    mask_vec = np.broadcast_to(mask.reshape(*mask.shape, 1), mask.shape + (3, ))

    # Now we fill the arrays to be sent to the renderer, also constructing vectors
    counter = 0
    for ind in range(len(to_render)):
        binned[counter][condition] = to_binning[counter][indices][condition]
        if scalar_layer[ind]:
            to_render[ind]["data"] = ma.masked_where(mask, binned[counter], copy=False)
            counter += 1
        else:
            for j in range(counter + 1, counter + 3):
                binned[j][condition] = to_binning[j][indices][condition]
            to_render[ind]["data"] = ma.masked_where(mask_vec,
                                                     np.array([
                                                         binned[counter].T,
                                                         binned[counter + 1].T,
                                                         binned[counter + 2].T
                                                     ]).T,
                                                     copy=False)
            counter += 3

    to_return = {"x": xcenters, "y": ycenters, "layers": to_render}
    if plot:
        # Render the map
        figure = render(x=xcenters, y=ycenters, data=to_render, ax=ax)
        figure["ax"].set_xlabel(dataset["amr"]["xyz"].x.label)
        figure["ax"].set_ylabel(dataset["amr"]["xyz"].y.label)
        if ax is None:
            figure["ax"].set_aspect("equal")

        # Add scatter layer
        if len(to_scatter) > 0:
            _add_scatter(datax=datax,
                         datay=datay,
                         to_scatter=to_scatter,
                         origin=origin,
                         datadx=datadx,
                         dir_vecs=dir_vecs,
                         dx=dx,
                         dy=dy,
                         ax=figure["ax"])

        figure["ax"].set_xlim(xmin, xmax)
        figure["ax"].set_ylim(ymin, ymax)

        to_return.update({"fig": figure["fig"], "ax": figure["ax"]})

    return Plot(**to_return)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)
import numpy as np
import os
from .reader import Reader, ReaderKind
from .. import config


class RtReader(Reader):
    def __init__(self):
        super().__init__(kind=ReaderKind.AMR)

    def initialize(self, meta, select):
        # Read the number of variables from the rt_file_descriptor.txt
        # and select the ones to be read if specified by user
        fname = os.path.join(meta["infile"], "rt_file_descriptor.txt")
        try:
            descriptor = np.loadtxt(fname, dtype=str, delimiter=",")
        except IOError:
            return

        for i in range(len(descriptor)):
            key = descriptor[i, 1].strip()
            read = True
            if isinstance(select, bool):
                read = select
            elif key in select:
                if isinstance(select[key], bool):
                    read = select[key]
            self.variables[key] = {
                "read": read,
                "type": descriptor[i, 2].strip(),
                "buffer": None,
                "pieces": {},
                "unit": config.get_unit(key, meta["unit_d"], meta["unit_l"],
                                        meta["unit_t"])
            }
        self.initialized = True

    def read_header(self, info):
        self.offsets["i"] += 5
        self.offsets["n"] += 6
        self.offsets["d"] += 1

    def read_domain_header(self):
        self.offsets['n'] += 2
        self.offsets['i'] += 2
